from ..utils import Scraper
from bs4 import BeautifulSoup
from collections import OrderedDict
from datetime import datetime
from pytz import timezone
from time import sleep
import re
import http.cookiejar


class UTSGDates:
    """A scraper for UTSG important dates."""

    @staticmethod
    def scrape(location='.', year=None, save=True):
        Scraper.logger.info('UTSGDates initialized.')

        docs = OrderedDict()

        for faculty in ArtSciDates, EngDates:
            dates = faculty.scrape(location, year=year, save=False)

            if dates is None:
                continue

            for date, doc in dates.items():
                if date not in docs:
                    docs[date] = OrderedDict([
                        ('date', date),
                        ('events', [])
                    ])

                docs[date]['events'].extend(doc['events'])

        if save:
            for date, doc in docs.items():
                Scraper.save_json(doc, location, date)

        Scraper.logger.info('UTSGDates completed.')
        return docs if not save else None


class ArtSciDates:
    """A scraper for important dates for UTSG Arts & Science.

    Data is retrieved from
    http://www.artsci.utoronto.ca/current/course/timetable/.
    """

    host = 'http://www.artsci.utoronto.ca/current/course/timetable/'

    @staticmethod
    def scrape(location='.', year=None, save=True):
        """Update the local JSON files for this scraper."""
        Scraper.logger.info('ArtSciDates initialized.')

        docs = OrderedDict()

        for endpoint in ArtSciDates.get_endpoints(year):
            headers = {'Referer': ArtSciDates.host}
            html = Scraper.get('%s%s' % (ArtSciDates.host, endpoint),
                               headers=headers,
                               max_attempts=3)

            if html is None:
                Scraper.logger.info('No data available for %s.' % endpoint.upper)
                continue

            soup = BeautifulSoup(html, 'html.parser')

            session = ArtSciDates.parse_session(soup)

            for tr in soup.find(class_='vertical listing').find_all('tr'):
                if tr.find('th'):
                    continue

                data = tr.find_all('td')

                start, end = ArtSciDates.parse_dates(data[0].text, session)

                descriptions = []
                for t in data[1].text.split(';\n'):
                    descriptions += ArtSciDates.normalize_text(t)

                events = []
                for description in descriptions:
                    events.append(OrderedDict([
                        ('end', end),
                        ('session', session.upper()),
                        ('campus', 'UTSG'),
                        ('description', description)
                    ]))

                doc = OrderedDict([
                    ('date', start),
                    ('events', events),
                ])

                if start not in docs:
                    docs[start] = doc
                else:
                    docs[start]['events'].extend(doc['events'])

        if save:
            for date, doc in docs.items():
                Scraper.save_json(doc, location, date)

        Scraper.logger.info('ArtSciDates completed.')
        return docs if not save else None

    @staticmethod
    def get_endpoints(year):
        try:
            date = datetime(year=year)
        except:
            year = None

        if year is None:
            year = datetime.now().strftime('%y')

        session = '%s%d_fw' % (year, int(year) + 1)

        endpoints = []

        headers = {'Referer': ArtSciDates.host}
        html = Scraper.get('%s%s' % (ArtSciDates.host, session),
                           headers=headers,
                           max_attempts=3)

        if html is None:
            return endpoints

        soup = BeautifulSoup(html, 'html.parser')

        for a in soup.find(id='portal-column-one').find_all('a'):
            if a.has_attr('title') and 'important dates' in a['title'].lower():
                endpoints.append(a['href'])

        return ['%s/%s' % (session, a.split('/')[-1]) for a in endpoints] + \
            ['20%s5/dates' % year]

    @staticmethod
    def parse_session(soup):
        session = ''
        if soup.find(id='parent-fieldname-title'):
            session = soup.find(id='parent-fieldname-title').text
            session = session.replace('Important Dates', '').replace(':', '')
        else:
            # TODO parse page title
            pass
        return session.strip()

    @staticmethod
    def parse_dates(date, session):

        def get_full_date(partial_date):
            """Convert a partial date of the form `B d` (e.g. November 8)
            to the form `B d Y` (e.g. November 8 2016)."""
            month, day = partial_date.split(' ')
            year = session[:4]
            return '%s %s %s' % (month, day, year)

        date = date.replace(' to ', '-').replace('(tentative)', '').strip()
        if '-' in date:
            # Date range (e.g. December 21 - January 4 or November 7-8)
            if ' - ' in date:
                date = date.split(' - ')
                start, end = get_full_date(date[0]), get_full_date(date[1])
            else:
                month, days = date.split(' ')
                days = days.split('-')

                start, end = get_full_date('%s %s' % (month, days[0])), \
                    get_full_date('%s %s' % (month, days[1]))
        else:
            start = end = get_full_date(date)

        return ArtSciDates.convert_date(start), ArtSciDates.convert_date(end)

    @staticmethod
    def convert_date(date):
        """Convert a date of form `B d Y` (eg. May 13 2016) to ISO-8601."""
        return datetime.strptime(date, '%B %d %Y').date().isoformat()

    @staticmethod
    def normalize_text(text):
        text = re.sub(r'\s\s+', ' ', text).strip()

        if text == '':
            return []

        if '\n' in text and text[-2:] != '\n':
            return text.split('\n')

        return [text]


class EngDates:
    """A scraper for important dates for UTSG Engineering.

    Data is retrieved from
    http://www.undergrad.engineering.utoronto.ca/About/Dates_Deadlines.htm.
    """

    host = 'http://www.undergrad.engineering.utoronto.ca/About/Dates_Deadlines.htm'

    cookies = http.cookiejar.CookieJar()

    form_data = {
        'viewstate': '__VIEWSTATE',
        'viewstate_generator': '__VIEWSTATEGENERATOR',
        'numerical_date': 'ctl02$ctlSelectedDate$hdnDateValueForQuestionnaireResponses',
        'textual_date': 'ctl02$ctlSelectedDate$txtDate'
    }

    @staticmethod
    def scrape(location='.', year=None, save=True):
        """Update the local JSON files for this scraper."""
        Scraper.logger.info('EngDates initialized.')

        year = year or datetime.now().year

        viewstate, viewstate_generator, numerical_date, textual_date = \
            EngDates.form_data.values()

        headers = {
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8',
            'Accept-Encoding': 'gzip, deflate',
            'Connection': 'keep-alive',
            'Content-Type': 'application/x-www-form-urlencoded',
            'Referer': EngDates.host,
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.94 Safari/537.36'
        }

        s = Scraper.s
        s.headers.update(headers)

        payload = {}
        payload[viewstate], payload[viewstate_generator] = \
            EngDates.get_viewstate(s)

        docs = OrderedDict()

        for m in range(1, 13):
            Scraper.logger.info('Scraping month %d' % m)

            month = datetime.strptime(str(m), '%m').strftime('%B')
            payload[textual_date] = '%s 15 %s' % (month, year)
            payload[numerical_date] = '%s.%s.15' % (year, str(m).zfill(2))

            attempts = 0

            html = s.post(EngDates.host,
                          data=payload,
                          cookies=EngDates.cookies).text.encode('utf-8') or ''
            soup = BeautifulSoup(html, 'html.parser')

            while attempts < 5 and soup.find('div', class_='error'):
                Scraper.logger.info('Attempt #%d' % (attempts + 1))

                html = s.post(EngDates.host,
                              data=payload,
                              cookies=EngDates.cookies).text.encode('utf-8') or ''
                soup = BeautifulSoup(html, 'html.parser')

                attempts += 1
                sleep(1)

            if not html or soup.find('div', class_='error'):
                Scraper.logger.info('Couldn\'t scrape month %d' % m)
                continue

            for tr in soup.find(id='ctl02_ctlCalendar').find_all('tr')[2:]:
                for td in tr.find_all('td'):
                    if not td.find('a') or not td.find('div').find('a'):
                        continue

                    start = end = EngDates.get_date(td.find('a')['title'], year)

                    session = '%d ENGINEERING' % year

                    events = []
                    for div in td.find_all('div'):
                        event = div.find('a')

                        events.append(OrderedDict([
                            ('end_date', end),
                            ('session', session),
                            ('campus', 'UTSG'),
                            ('description', event.text)
                        ]))

                    if start not in docs:
                        docs[start] = OrderedDict([
                            ('date', start),
                            ('events', events)
                        ])
                    else:
                        docs[start]['events'].extend(events)

        if save:
            for date, doc in docs.items():
                Scraper.save_json(doc, location, date)

        Scraper.logger.info('EngDates completed.')
        return docs if not save else None

    @staticmethod
    def get_date(date, year):
        """Return a IS0 8601 date from a date string of the form `M d`"""
        date = '%s %s' % (year, date)
        return datetime.strptime(date, '%Y %B %d').date().isoformat()

    @staticmethod
    def get_viewstate(s):
        headers = {'Referer': EngDates.host}
        html = s.get(EngDates.host, headers=headers)
        soup = BeautifulSoup(html.content, 'html.parser')

        return soup.find(id='__VIEWSTATE')['value'],\
            soup.find(id='__VIEWSTATEGENERATOR')['value']



from datetime import datetime


def convert_time(t):
    t = [int(x) for x in t.split(':')]

    converted = 0
    for i in range(min(len(t), 3)):
        converted += t[i] * (60 ** (2-i))
    return converted


def get_period(d):
    def get_date(month, date, year):
        month = 'jun' if month == 'june' else month
        return datetime.strptime('%s%s%d' % (year, month, date), '%Y%b%d')

    d = datetime.strptime(d, '%Y-%m-%d')
    year, month = d.year, None

    for m, ld in (('dec', 31), ('apr', 30), ('june', 30), ('aug', 31)):
        if get_date(m, 1, year) <= d <= get_date(m, ld, year):
            month = m
            break

    if month:
        return '%s%s' % (month.upper(), str(year)[2:])


def get_course_id(course_code, date):
        d = datetime.strptime(date, '%Y-%m-%d')

        month, year = d.strftime('%b').lower(), d.year
        month = 'june' if month == 'jun' else month

        endings = {
            'dec': {
                'F': '%d9' % year,
                'Y': '%d9' % (year - 1)
            },
            'apr': {
                'S': '%d1' % year,
                'Y': '%d9' % (year - 1)
            },
            'june': {
                'F': '%d5F' % year,
                'Y': '%d5' % year
            },
            'aug': {
                'S': '%d5S' % year,
                'Y': '%d5' % year
            }
        }

        season = course_code[-1]
        period = get_period(date)

        exam_id = course_id = None

        if month in endings and season in endings[month]:
            course_id = '%s%s' % (course_code, endings[month][season])
            exam_id = '%s%s' % (course_id, period)

        return exam_id, course_id



from ..utils import Scraper
from bs4 import BeautifulSoup
from collections import OrderedDict
import re
import tidylib
from math import floor
import json


class UTSGTimetable:

    host = 'https://timetable.iit.artsci.utoronto.ca/api'
    day_map = {
        'MO': 'MONDAY',
        'TU': 'TUESDAY',
        'WE': 'WEDNESDAY',
        'TH': 'THURSDAY',
        'FR': 'FRIDAY',
        'SA': 'SATURDAY',
        'SU': 'SUNDAY'
    }

    @staticmethod
    def scrape(location='.'):
        Scraper.logger.info('UTSGTimetable initialized.')

        orgs = UTSGTimetable.get_orgs()
        for org in orgs:
            Scraper.logger.info('Scraping %s.' % org)

            data = UTSGTimetable.search(org)

            if not data:
                continue

            for c in data.keys():
                x = data[c]

                course_id = '%s%s%s' % (x['code'], x['section'], x['session'])
                course_code = '%s%s' % (x['code'], x['section'])
                course_name = x['courseTitle']
                description = BeautifulSoup(x['courseDescription'], 'html.parser').text
                division = 'Faculty of Arts and Science'
                department = x['orgName']
                prerequisites = x['prerequisite']
                exclusions = x['exclusion']

                m = re.search('(?:[^\d]*)(\d+)', x['code'])
                level = int(floor(int(m.group(1)) / 100.0)) * 100

                campus = 'UTSG'

                year = x['session'][:4]
                month = x['session'][4:]
                term = ''
                if month == '9':
                    term = '%s Fall' % year
                elif month == '1':
                    term = '%s Winter' % year
                elif month == '5':
                    term = '%s Summer Y' % year
                elif month == '5F':
                    term = '%s Summer F' % year
                elif month == '5S':
                    term = '%s Summer S' % year

                breadths = []
                for ch in x['breadthCategories']:
                    if ch in '12345':
                        breadths.append(int(ch))
                breadths = sorted(breadths)

                sections = []

                meetings = x['meetings']
                if meetings:
                    for meeting in meetings.keys():
                        y = meetings[meeting]

                        code = '%s%s' % (y['teachingMethod'][:1], y['sectionNumber'])

                        instructors = []
                        if y['instructors']:
                            for instructor_id in y['instructors'].keys():
                                instructor = y['instructors'][instructor_id]
                                formatted = '%s %s' % (
                                    instructor['firstName'][:1],
                                    instructor['lastName']
                                )
                                formatted = formatted.strip()
                                if len(formatted) == 0:
                                    continue
                                instructors.append(formatted)


                        size = 0
                        if 'enrollmentCapacity' in y.keys():
                            size = y['enrollmentCapacity']
                            if size:
                                if len(size) > 0:
                                    size = int(size)
                                else:
                                    size = 0
                            else:
                                size = 0

                        # TODO: they haven't added this yet
                        enrolment = 0

                        times = []
                        if y['schedule']:
                            for time_id in y['schedule'].keys():
                                z = y['schedule'][time_id]

                                if z['meetingStartTime'] == None:
                                    continue

                                day = ''
                                if z['meetingDay'] in UTSGTimetable.day_map.keys():
                                    day = UTSGTimetable.day_map[z['meetingDay']]

                                startTime = z['meetingStartTime'].split(':')
                                start = (60 * 60 * int(startTime[0])) + (int(startTime[1]) * 60)

                                endTime = z['meetingEndTime'].split(':')
                                end = (60 * 60 * int(endTime[0])) + (int(endTime[1]) * 60)

                                times.append(OrderedDict([
                                    ("day", day),
                                    ("start", start),
                                    ("end", end),
                                    ("duration", end - start),
                                    ("location", '')
                                ]))

                        sections.append(OrderedDict([
                            ("code", code),
                            ("instructors", instructors),
                            ("times", times),
                            ("size", size),
                            ("enrolment", enrolment)
                        ]))

                course = OrderedDict([
                    ("id", course_id),
                    ("code", course_code),
                    ("name", course_name),
                    ("description", description),
                    ("division", division),
                    ("department", department),
                    ("prerequisites", prerequisites),
                    ("exclusions", exclusions),
                    ("level", level),
                    ("campus", campus),
                    ("term", term),
                    ("breadths", breadths),
                    ("meeting_sections", sections)
                ])

                Scraper.save_json(course, location, course_id)

        Scraper.logger.info('UTSGTimetable completed.')

    @staticmethod
    def get_orgs():
        data = Scraper.get('%s/orgs' % UTSGTimetable.host, json=True)
        if 'orgs' in data.keys():
            return list(data['orgs'].keys())
    @staticmethod
    def search(org):
        data = Scraper.get('%s/courses?org=%s' % (UTSGTimetable.host, org),
            json=True, timeout=60)
        if data:
            return data



from ..utils import Scraper
from bs4 import BeautifulSoup, NavigableString, Comment
from collections import OrderedDict
from datetime import datetime, date
from urllib.parse import urlencode
import re
import urllib.parse as urlparse


class Events:
    """A scraper for Events at the University of Toronto."""
    host = 'https://www.events.utoronto.ca/'

    campuses_tags = {'St. George': 'UTSG', 'U of T Mississauga': 'UTM',
                     'U of T Scarborough': 'UTSC'}

    @staticmethod
    def scrape(location='.'):
        Scraper.logger.info('Events initialized.')
        Scraper.ensure_location(location)

        for event in Events.get_events_list():
            doc = Events.get_event_doc(event[0], event[1])
            Scraper.save_json(doc, location, doc['id'])

        Scraper.logger.info('Events completed.')

    @staticmethod
    def get_events_list():
        page_index_url = Events.host + 'index.php'
        url_parts = list(urlparse.urlparse(page_index_url))
        events_links, events_dates = [], []
        paging_index = 1
        events_count = 10

        while events_count == 10:
            params = {
                'p': paging_index
            }
            url_parts[4] = urlencode(params)
            paging_index += 1
            html = Scraper.get(urlparse.urlunparse(url_parts))
            soup = BeautifulSoup(html, 'html.parser')
            events_dom_arr = soup.select('#results')[0].find_all('li')
            events_count = len(events_dom_arr)
            events_links += list(map(lambda e: e.a['href'], events_dom_arr))
            events_dates += list(map(lambda e: e.find('p').text.split(' : ')[1].split(', ')[0], events_dom_arr))

        return zip(events_links, events_dates)

    @staticmethod
    def convert_time(time_str):
        hour_tks = time_str[:-2].split(':')
        meridiem = time_str[-2:]
        hours = int(hour_tks[0])
        minutes = 0
        if len(hour_tks) > 1:
            minutes = int(hour_tks[1])
        if (meridiem == 'pm'):
            if (int(hours) != 12):
                hours = int(hours) + 12
        posix_from_midnight = hours*60*60 + minutes*60
        return posix_from_midnight

    @staticmethod
    def normalize_text_sections(div):
        paragraph = ''
        for content in div.contents:
            text = ''
            if type(content) == NavigableString:
                text = content
            elif type(content) == Comment:
                pass
            elif content.name == 'li':
                text = content.text
            else:
                text = content.text
            text = text.strip()
            paragraph += text.strip() + ' '
        paragraph = paragraph.strip()
        paragraph = paragraph.replace('\r', '')
        paragraph = paragraph.replace('\n', ', ')
        paragraph = paragraph.replace('  ', ' ')
        paragraph = paragraph.strip()
        return paragraph

    @staticmethod
    def get_event_doc(url_tail, event_date):
        event_url = Events.host + url_tail
        html = Scraper.get(event_url)
        url_parts = list(urlparse.urlparse(event_url))
        query = dict(urlparse.parse_qsl(url_parts[4]))
        soup = BeautifulSoup(html, 'html.parser')

        event_id = query['eventid']
        event_title = soup.select('.eventTitle')[0].text.strip()

        date_arr = event_date.split(' - ')

        start_date = date_arr[0].strip()
        end_date = start_date if len(date_arr) == 1 else date_arr[1].strip()

        if start_date.count(' ') == 1:
            # year not in start date
            start_date = '%s %s' % (start_date, end_date[-4:])

        start_date = datetime.strptime(start_date, '%b %d %Y')
        end_date = datetime.strptime(end_date, '%b %d %Y')

        event_start_date = start_date.date().isoformat()
        event_end_date = end_date.date().isoformat()

        raw_time = soup.select('.date')[0].text.split(',')

        time_arr = re.split(' - | ', raw_time[1].strip())

        # Some of the strings are misformed and gives an extra empty space
        time_arr = list(filter(None, time_arr))

        event_start_str = time_arr[0]
        event_end_str = time_arr[-2] + time_arr[-1]
        if (len(time_arr) == 3):
            event_start_str += time_arr[-1]
        else:
            event_start_str += time_arr[1]

        event_start_time = Events.convert_time(event_start_str)
        event_end_time = Events.convert_time(event_end_str)

        event_duration = event_end_time - event_start_time

        evt_bar = soup.select('#evt_bar')[0]
        event_url = evt_bar.select('dd')[1].a['href']
        event_price = evt_bar.select('dl')[1].dd.text

        event_campus = ''
        if evt_bar.select('dd')[0].b is not None:
            campus_full_name = evt_bar.select('dd')[0].b.text
            event_campus = Events.campuses_tags[campus_full_name]

        event_address = ''
        address_block = evt_bar.select('dd')[0]
        if address_block.a is not None:
            address_block = address_block.a
        event_address = Events.normalize_text_sections(address_block)

        event_audiences = list(
            map(lambda a: a.text, evt_bar.select(
                'dl')[1].select('dd')[1].select('a')))

        soup.select('.eventTitle')[0].extract()
        soup.select('.date')[0].extract()
        evt_bar.extract()
        soup.select('#cal_bar')[0].extract()
        event_description = Events.normalize_text_sections(
            soup.select('#content')[0])

        doc = OrderedDict([
            ('id', event_id),
            ('title', event_title),
            ('start_date', event_start_date),
            ('end_date', event_end_date),
            ('start_time', event_start_time),
            ('end_time', event_end_time),
            ('duration', event_duration),
            ('url', event_url),
            ('description', event_description),
            ('admission_price', event_price),
            ('campus', event_campus),
            ('location', event_address),
            ('audiences', event_audiences)
        ])
        return doc



from ..utils import Scraper
from .exams_helpers import *
from bs4 import BeautifulSoup
from collections import OrderedDict
from datetime import datetime
from pytz import timezone


class UTSCExams:
    """A scraper for UTSC exams."""

    host = 'http://www.utsc.utoronto.ca/registrar/examination-schedule'

    @staticmethod
    def scrape(location='.'):
        Scraper.logger.info('UTSCExams initialized.')

        exams = OrderedDict()

        headers = {
            'Referer': UTSCExams.host
        }
        html = Scraper.get('%s' % UTSCExams.host, headers=headers)
        soup = BeautifulSoup(html, 'html.parser')

        for table in soup.find_all('table', class_='views-table'):
            for tr in table.find_all('tr')[1:]:
                data = [x.text.strip() for x in tr.find_all('td')]

                course_code, lecture_code = data[0], None
                if ' ' in course_code:
                    course_code, lecture_code = course_code.split(' ')

                date = data[1]

                exam_id, course_id = get_course_id(course_code, date)

                period = get_period(date)

                if not exam_id or not period:
                    continue

                start = convert_time(data[2])
                end = convert_time(data[3])
                duration = end - start

                location_ = data[4]

                doc = OrderedDict([
                    ('id', exam_id),
                    ('course_id', course_id),
                    ('course_code', course_code),
                    ('campus', 'UTSC'),
                    ('period', period),
                    ('date', date),
                    ('start_time', start),
                    ('end_time', end),
                    ('duration', duration),
                    ('sections', [])
                ])

                if exam_id not in exams:
                    exams[exam_id] = doc

                exams[exam_id]['sections'].append(OrderedDict([
                    ('lecture_code', lecture_code or ''),
                    ('exam_section', ''),
                    ('location', location_)
                ]))

        for id_, doc in exams.items():
            Scraper.save_json(doc, location, id_)

        Scraper.logger.info('UTSCExams completed.')



from ..utils import Scraper
from .athletics_helpers import *
from bs4 import BeautifulSoup
from datetime import datetime
from collections import OrderedDict
import json
import requests


class UTSCAthletics:
    """A scraper for the UTSC athletics schedule.

    The schedule is located at http://www.utsc.utoronto.ca/athletics
    """

    host = 'http://www.utsc.utoronto.ca/athletics/calendar-node-field-date-time/month/'

    @staticmethod
    def scrape(location='.', month=None, save=True):
        """Update the local JSON files for this scraper."""
        month = month or get_current_month()

        Scraper.logger.info('UTSCAthletics initialized.')
        html = Scraper.get('%s%s' % (UTSCAthletics.host, month))
        soup = BeautifulSoup(html, 'html.parser')

        athletics = OrderedDict()

        calendar = soup.find('div', class_='month-view')
        for tr in calendar.find_all('tr', class_='single-day'):
            for td in tr.find_all('td'):
                date = td.get('data-date')
                id_ = get_campus_id(date, 'SC')

                if not is_date_in_month(date, month):
                    continue

                events = []

                for item in td.find(class_='inner').find_all(class_='item'):
                    title = item.find(class_='views-field-title').text.strip()

                    location_ = item.find(class_='views-field-field-location]')

                    if location_.text.strip() == '':
                        location_ = list(location_.next_siblings)[1]

                    location_ = location_.text.strip()

                    start = convert_time(item.find(class_='date-display-start').get('content'))
                    end = convert_time(item.find(class_='date-display-end').get('content'))

                    duration = end - start

                    events.append(OrderedDict([
                        ('title', title.replace('/ ', '/')),
                        ('campus', 'UTSC'),
                        ('location', location_),
                        ('building_id', '208'),
                        ('start_time', start),
                        ('end_time', end),
                        ('duration', duration)
                    ]))

                athletics[date] = OrderedDict([
                    ('date', date),
                    ('events', events)
                ])

        if save:
            for id_, doc in athletics.items():
                Scraper.save_json(doc, location, id_)

        Scraper.logger.info('UTSCAthletics completed.')

        return athletics if not save else None



from ..utils import Scraper
from .utsg import UTSGDates
from .utm import UTMDates

from collections import OrderedDict


class Dates:

    @staticmethod
    def scrape(location='.', year=None):
        Scraper.logger.info('Dates initialized.')

        docs = OrderedDict()

        for campus in UTSGDates, UTMDates:
            dates = campus.scrape(location, year=year, save=False)

            if dates is None:
                continue

            for date, doc in dates.items():
                if date not in docs:
                    docs[date] = OrderedDict([
                        ('date', date),
                        ('events', [])
                    ])
                docs[date]['events'].extend(doc['events'])

        for date, doc in docs.items():
            Scraper.save_json(doc, location, date)

        Scraper.logger.info('Dates completed.')



from ..utils import Scraper
from bs4 import BeautifulSoup, NavigableString, Comment
from collections import OrderedDict


class Libraries:
    """A scraper for the Libraries at the University of Toronto."""
    host = 'https://onesearch.library.utoronto.ca/'

    @staticmethod
    def scrape(location='.'):
        Scraper.logger.info('Libraries initialized.')
        Scraper.ensure_location(location)

        for library_link in Libraries.get_library_link():
            doc = Libraries.get_library_doc(library_link)
            if (doc is not None):
                Scraper.save_json(
                    doc, location, library_link.split('/')[-1].upper())
                Scraper.logger.info(
                    'Scraped ' + library_link.split('/')[-1].upper() + '.')
            else:
                # Not a real library page
                Scraper.logger.info(
                    'Skipped ' + library_link.split('/')[-1].upper() + '.')

        Scraper.logger.info('Libraries completed.')

    @staticmethod
    def get_library_link():
        vist_libraries_url = Libraries.host + 'visit'
        html = Scraper.get(vist_libraries_url)
        soup = BeautifulSoup(html, 'html.parser')
        list_obj_arr = soup.select('.view-list-of-libraries')[1].select(
            '.view-content')[0].select('.views-row')
        library_links = [l.a['href'] for l in list_obj_arr]
        return library_links

    @staticmethod
    def normalize_text_sections(div):
        paragraph = ''
        for content in div.contents:
            text = ''
            if type(content) == NavigableString:
                text = content
            elif type(content) == Comment:
                pass
            elif content.name == 'li':
                text = content.text
            else:
                text = content.text
            text = text.strip()
            paragraph += text.strip() + ' '
        paragraph = paragraph.strip()
        paragraph = paragraph.replace('\r', '')
        paragraph = paragraph.replace('\n', ', ')
        paragraph = paragraph.strip()
        return paragraph

    @staticmethod
    def convert_time(time_str):
        end = len(time_str)
        if 'am' in time_str:
            end = time_str.index('am') + 2
        elif 'pm' in time_str:
            end = time_str.index('pm') + 2
        time_str = time_str[:end]
        hour_tks = time_str[:-2].split(':')
        meridiem = time_str[-2:]
        hours = int(hour_tks[0])
        minutes = 0
        if len(hour_tks) > 1:
            minutes = int(hour_tks[1])
        if (meridiem == 'pm'):
            if (int(hours) != 12):
                hours = int(hours) + 12
        posix_from_midnight = hours*60*60 + minutes*60
        return posix_from_midnight

    @staticmethod
    def get_library_hours(calendar_link):
        weekdays = ['sunday', 'monday', 'tuesday', 'wednesday',
                    'thursday', 'friday', 'saturday']
        html = Scraper.get(calendar_link, max_attempts=1)
        week = ['Closed'] * 7
        if html is not None:
            soup = BeautifulSoup(html, 'html.parser')
            if soup.select('.calendar-wrapper') != []:
                week = soup.select('.calendar-wrapper')[0].select(
                    '.start-week')[1].select('.start-day')
                week = list(map(Libraries.normalize_text_sections, week))
        hours = OrderedDict()
        week = [day[1:].strip() for day in week]
        for day in range(len(weekdays)):
            hour = week[day]
            closed = not hour.startswith('Open')
            opening_hour = 0
            closing_hour = 0
            if not closed:
                hour_tks = hour.replace('Open:', '').split('-')
                opening_hour = Libraries.convert_time(hour_tks[0])
                closing_hour = Libraries.convert_time(hour_tks[1])
            hours[weekdays[day]] = OrderedDict([
                ('closed', closed),
                ('open', opening_hour),
                ('close', closing_hour)
            ])
        return hours

    @staticmethod
    def get_library_doc(url_tail):
        _id = url_tail.split('/')[-1].upper()

        library_url = Libraries.host + url_tail
        html = Scraper.get(library_url)
        soup = BeautifulSoup(html, 'html.parser')
        main_content = soup.select('#content-inner')[0].select('.library-info')

        if main_content == []:
            return None
        else:
            main_content = main_content[0]

        library_name = main_content.h1.extract().text

        library_image = ''
        if main_content.img is not None:
            library_image = main_content.img.extract()['src']

        for content in main_content.select('.field-content'):
            content.extract()

        library_website = main_content.a.extract()['href']

        library_hours_link = main_content.a['href']
        library_hours = Libraries.get_library_hours(library_hours_link)

        library_address = ''
        if main_content.select('.library-address') != []:
            library_address = Libraries.normalize_text_sections(
                main_content.select('.library-address')[0].extract())

        library_phone = ''
        if main_content.select('.phone') != []:
            library_phone = main_content.select(
                '.phone')[0].extract().text.strip()

        library_info = main_content.select('.library-info-text')[0].extract()
        library_info_titles = [s.text for s in library_info.select('h2')]
        library_info_texts = []

        for content in library_info.contents:
            if type(content) != NavigableString:
                if content.name in ['p', 'ul', 'ol']:
                    library_info_texts.append(content)
        library_info_texts = list(
            map(Libraries.normalize_text_sections, library_info_texts))

        library_about = ''
        library_collection_strengths = ''
        library_how_to_access = ''

        for i in range(len(library_info_titles)):
            if library_info_titles[i] == 'About the library':
                library_about = library_info_texts[i]
            elif library_info_titles[i] == 'Collection strengths':
                library_collection_strengths = library_info_texts[i].replace(
                    '  ', ', ')
            elif library_info_titles[i] == 'How to access':
                library_how_to_access = library_info_texts[i]

        doc = OrderedDict([
            ('id', _id),
            ('name', library_name),
            ('image', library_image),
            ('website', library_website),
            ('address', library_address),
            ('phone', library_phone),
            ('about', library_about),
            ('collection_strengths', library_collection_strengths),
            ('access', library_how_to_access),
            ('hours', library_hours)
        ])
        return doc



from ..utils import Scraper
from bs4 import BeautifulSoup
from collections import OrderedDict
import json
import os
import requests
import datetime


class UTMDates:
    '''Scraper for Important dates from UTM calendar found at https://www.utm.utoronto.ca/registrar/important-dates
        '''

    link = 'http://m.utm.utoronto.ca/importantDates.php?mode=full&session={0}{1}&header='
    sessionNumber = [5, 9]
    @staticmethod
    def scrape(location='.', year=None, save=True):  # scrapes most current sessions by default

        year = year or datetime.datetime.now().year

        currentSession = "{0} SUMMER"
        calendar = OrderedDict()
        Scraper.logger.info('UTMDates initialized.')
        for session in UTMDates.sessionNumber:
            html = Scraper.get(UTMDates.link.format(year, session))
            soup = BeautifulSoup(html, 'html.parser')
            content = soup.find('div', class_='content')
            dates = content.find_all('div', class_='title')
            i = 0
            currentDate = dates[i]
            while(i < len(dates)):
                date = dates[i].text
                events = []
                while (currentDate == dates[i]):
                    info = dates[i].find_next('div', class_='info')
                    description = info.text
                    eventStartEnd = date.split('-')  # splits event dates over a period
                    eventStart = UTMDates.convert_date(eventStartEnd[0].strip())
                    if len(eventStartEnd) > 1:
                        eventEnd = UTMDates.convert_date(eventStartEnd[1].strip())
                    else:
                        eventEnd = eventStart

                    events.append(OrderedDict([
                            ('end_date', eventEnd),
                            ('session', currentSession.format(UTMDates.get_year_from(eventStart))),
                            ('campus', 'UTM'),
                            ('description', description)
                        ]))
                    i += 1
                    if(i >= len(dates)):
                        break
                calendar[eventStart] = OrderedDict([
                    ('date', eventStart),
                    ('events', events)
                ])
                if(i < len(dates)):
                    currentDate = dates[i]
            currentSession = "{0} FALL/WINTER"

        if save:
            for date, info in calendar.items():
                Scraper.save_json(info, location, UTMDates.convert_date(date))

        Scraper.logger.info('UTMDates completed.')
        return calendar if not save else None

    @staticmethod
    def convert_date(date):
        splitDate = date.split(' ')
        year = splitDate[2]
        day = splitDate[1].strip(',')
        month = datetime.datetime.strptime(splitDate[0], '%B').strftime('%m')
        return("{0}-{1}-{2}".format(year, month, day.zfill(2)))

    @staticmethod
    def get_year_from(date):
        splitDate = date.split('-')
        return splitDate[0]



from datetime import datetime


def get_current_month():
    """Return current month."""
    now = datetime.now()
    return '%s-%s' % (now.year, now.month)


def get_campus_id(d, campus):
    """Return campus id, made up of date and specifier (one of SG, M, SC)."""
    d = datetime.strptime(d, '%Y-%m-%d')
    return '%s%s' % (str(d.day).zfill(2), campus)


def is_date_in_month(d, m):
    """Determine if the given date is in the given month."""
    d, m = datetime.strptime(d, '%Y-%m-%d'), datetime.strptime(m, '%Y-%m')
    return d.month == m.month


def convert_time(dt):
    """Convert datetime from ISO 8601 format to seconds since midnight."""
    dt = datetime.strptime(dt[:19], '%Y-%m-%dT%H:%M:%S')
    return dt.hour * 60 * 60 + dt.minute * 60



from ..utils import Scraper
from .exams_helpers import *
from bs4 import BeautifulSoup
from collections import OrderedDict
from datetime import datetime
from pytz import timezone


class UTSGExams:
    """A scraper for UTSG exams."""

    @staticmethod
    def scrape(location='.', year=None):
        """Update the local JSON files for this scraper."""

        Scraper.logger.info('UTSGExams initialized.')

        for faculty in ArtSciExams, EngExams:
            exams = faculty.scrape(location=location, year=year, save=False)
            if exams is None:
                continue
            for id_, doc in exams.items():
                Scraper.save_json(doc, location, id_)

        Scraper.logger.info('UTSGExams completed.')


class ArtSciExams:
    """A scraper for Arts & Science exams.

    Data is scraped from http://www.artsci.utoronto.ca/current/exams/
    """

    host = 'http://www.artsci.utoronto.ca/current/exams/'

    @staticmethod
    def scrape(location='.', year=None, save=True):
        """Update the local JSON files for this scraper."""

        Scraper.logger.info('ArtSciExams initialized.')

        exams = OrderedDict()

        for p in ArtSciExams.get_exam_periods(year):
            Scraper.logger.info('Scraping %s exams.' % p.upper())

            headers = {
                'Referer': ArtSciExams.host
            }
            html = Scraper.get('%s%s' % (ArtSciExams.host, p),
                               headers=headers,
                               max_attempts=3)

            try:
                soup = BeautifulSoup(html, 'html.parser')
            except TypeError:
                soup = None

            if not (html and soup and soup.find(class_='vertical listing')):
                # no exam data available
                Scraper.logger.info('No %s exams.' % p.upper())
                continue

            rows = soup.find('table', class_='vertical listing').find_all('tr')
            for row in rows[1:]:
                data = [x.text.strip() for x in row.find_all('td')]

                exam_id, course_id, course_code = \
                    ArtSciExams.parse_course_info(p, data[0])

                if exam_id is None:
                    continue

                section = data[1]

                lecture_section = exam_section = None

                if '  ' in section:
                    lecture_section, exam_section = section.split('  ')
                elif '-' in section:
                    exam_section = section
                else:
                    lecture_section = section

                location_ = data[4]

                date = ArtSciExams.parse_date(data[2], p[-2:]) or ''
                start, end = ArtSciExams.parse_time(data[3], date) or (0, 0)
                duration = end - start

                doc = OrderedDict([
                    ('id', exam_id),
                    ('course_id', course_id),
                    ('course_code', course_code),
                    ('campus', 'UTSG'),
                    ('period', p.upper()),
                    ('date', date),
                    ('start_time', start),
                    ('end_time', end),
                    ('duration', duration),
                    ('sections', [])
                ])

                if exam_id not in exams:
                    exams[exam_id] = doc

                exams[exam_id]['sections'].append(OrderedDict([
                    ('lecture_code', lecture_section or ''),
                    ('exam_section', exam_section or ''),
                    ('location', location_)
                ]))

        if save:
            for id_, doc in exams.items():
                Scraper.save_json(doc, location, id_)

        Scraper.logger.info('ArtSciExams completed.')
        return exams if not save else None

    @staticmethod
    def parse_course_info(period, course_code):

        month, year, season = period[:-2], period[-2:], course_code[-1]

        endings = {
            'dec': {
                'F': '%s9' % str(int(year)),
                'Y': '%s9' % str(int(year) - 1)
            },
            'apr': {
                'S': '%s1' % str(year),
                'Y': '%s9' % str(int(year) - 1)
            },
            'june': {
                'F': '%s5F' % str(year),
                'Y': '%s5' % str(year)
            },
            'aug': {
                'S': '%s5S' % str(year),
                'Y': '%s5' % str(year)
            }
        }

        exam_id = course_id = None
        if month in endings and season in endings[month]:
            course_id = '%s20%s' % (course_code, endings[month][season])
            exam_id = '%s%s' % (course_id, period.upper())

        return exam_id, course_id, course_code

    @staticmethod
    def parse_date(date, year):
        """Convert date of form `D DD MMM` to ISO 8601 format."""
        if date.count(' ') == 2:
            return datetime.strptime('%s %s' % (date, year),
                                     '%a %d %b %y').date().isoformat()

    @staticmethod
    def parse_time(time, d):
        """Convert time from `pd HH:MM - HH:MM` to start & end datetime."""

        def convert_time(t, is_pm=False):
            """Convert time from `HH:MM` to an ISO 8601 datetime."""
            h, m = [int(x) for x in t.split(':')]
            h += 12 if is_pm and h != 12 else 0
            return (h * 60 * 60) + (m * 60)

        time = list(filter(None, time.replace('-', '').split(' ')))
        if len(time) == 3:
            period, start, end = time
            after_12 = period == 'PM' or period == 'EV'

            return convert_time(start, after_12), convert_time(end, after_12)

    @staticmethod
    def get_exam_periods(year):
        if not year:
            year = datetime.today().year

        periods = []
        for m in ('dec', 'apr', 'june', 'aug'):
            y = year if not m == 'dec' else int(year) - 1
            periods.append('%s%s' % (m, str(y)[2:]))

        return periods


class EngExams:
    """A scraper for Engineering exams.

    Data is scraped from http://www.apsc.utoronto.ca/timetable/fes.aspx
    """

    host = 'http://www.apsc.utoronto.ca/timetable/fes.aspx'

    @staticmethod
    def scrape(location='.', year=None, save=True):
        """Update the local JSON files for this scraper."""

        Scraper.logger.info('EngExams initialized.')

        exams = OrderedDict()

        headers = {
            'Referer': EngExams.host
        }
        html = Scraper.get(EngExams.host, headers=headers, max_attempts=3)
        soup = BeautifulSoup(html, 'html.parser')

        if soup is None:
            return

        for tr in soup.find('table', id='DataList1').find_all('tr'):
            for td in tr.find_all('td'):
                entry = td.find('div', id='logo')

                if entry is None:
                    continue

                info = entry.find('div')

                date, time = [br.next_sibling.strip()
                              for br in info.find_all('br')[:2]]

                date = datetime.strptime(date.split(':')[-1].strip(),
                                         '%b %d, %Y').date().isoformat()

                time = time.strip().split(':')
                hour = int(time[1])
                minute, meridiem = time[2].split(' ')

                hour += 12 if meridiem == 'PM' and hour != 12 else 0

                # No end times, using 2.5h for duration per
                # http://www.undergrad.engineering.utoronto.ca/Office_of_the_Registrar/Examinations/Schedules_Locations.htm
                start = hour * 60 * 60 + int(minute) * 60
                duration = 2 * 60 * 60 + 30 * 60
                end = start + duration

                period = get_period(date)

                exam_id, course_id, course_code = \
                    EngExams.get_course_info(info.find('strong').text.strip(), period)

                locations = entry.find('table', class_='xx')

                exam_sections = []
                for tr in locations.find_all('tr')[1:]:
                    location, range = [td.text.strip() for td in tr.find_all('td')[:2]]

                    exam_sections.append(OrderedDict([
                        ('lecture_code', ''),
                        ('exam_section', range),
                        ('location', location.replace('-', ' '))
                    ]))

                exams[exam_id] = OrderedDict([
                    ('id', exam_id),
                    ('course_id', course_id),
                    ('course_code', course_code),
                    ('campus', 'UTSG'),
                    ('period', period),
                    ('date', date),
                    ('start_time', start),
                    ('end_time', end),
                    ('duration', duration),
                    ('sections', exam_sections)
                ])

        if save:
            for id_, doc in exams.items():
                Scraper.save_json(doc, location, id_)

        Scraper.logger.info('EngExams completed.')
        return exams if not save else None

    @staticmethod
    def get_course_info(course, period):
        endings = {
            'dec': {'season': 'F', 'month': '9'},
            'apr': {'season': 'S', 'month': '1'},
            'june': {'season': 'F', 'month': '5F'},
            'aug': {'season': 'S', 'month': '5S'}
        }

        month, year = period[:-2].lower(), period[-2:]
        exam_id = course_id = course_code = None
        if month in endings:
            course_code = '%s%s' % (course, endings[month]['season'])
            course_id = '%s20%s%s' % (course_code, year, endings[month]['month'])
            exam_id = '%s%s' % (course_id, period)
        return exam_id, course_id, course_code



from ..utils import Scraper
from .utsg import UTSGAthletics
from .utm import UTMAthletics
from .utsc import UTSCAthletics

from collections import OrderedDict


class Athletics:
    @staticmethod
    def scrape(location='.', month=None):
        Scraper.logger.info('Athletics initialized.')

        docs = OrderedDict()

        for campus in UTSGAthletics, UTMAthletics, UTSCAthletics:
            athletics = campus.scrape(location, month=month, save=False)

            if athletics is None:
                continue

            for date, data in athletics.items():
                if date not in docs:
                    docs[date] = OrderedDict([
                        ('date', date),
                        ('events', [])
                    ])
                docs[date]['events'].extend(data['events'])

        for date, doc in docs.items():
            Scraper.save_json(doc, location, date)

        Scraper.logger.info('Athletics completed.')
